/* tree' utility for printing dir content as a tree.
(c) 2002,2003,2004
Stanislav Ievlev <inger@altlinux.org>
License: GPL
*/

#define _GNU_SOURCE
#include <assert.h>
#include <wchar.h>
#include <wctype.h>

#include "tree.h"

/*force or not colorization*/
int     is_colorize = 0;
int     force_color = 0;
int     force_no_color = 0;

/*depth of tree to display*/
long    max_level = -1;

/*TODO:plain (like find) mode*/

/*int     is_plain = 0;*/

/*filter or not filter*/
int     is_filter = 0;

/*display filter*/
char   *filter_type = NULL;
int     filter[S_IFMT >> 12];	/*I have to use this scheme, 'cause mode_t masks intersects */
char   *filter_name = NULL;
regex_t re;

graphics *gr = &sym_ascii;

extern const char *__progname;

/*restore terminal colors to default values*/
static void
restore_default_colors ()
{
	colorize ("", NORMAL);
}

static void
restore_default_colors_handler (int sig)
{
	restore_default_colors ();
	_exit (128 + sig);
}

static void
sigtstp_handler (int sig)
{
	assert(sig == SIGTSTP);
	signal (sig, sigtstp_handler);

	restore_default_colors ();

	raise (SIGSTOP);
}

/*fill tokens with current values*/
void
process_token (const char *id, const char *token)
{
#if 0
	printf ("id=%s value=%s\n", id, token);
#endif
	if (!strcmp (id, "di"))
	{
		strncpy (tokens[DIRECTORY], token, MAX_TOKEN);
	} else if (!strcmp (id, "fi"))
	{
		strncpy (tokens[REGULAR], token, MAX_TOKEN);
	} else if (!strcmp (id, "ex"))
	{
		strncpy (tokens[EXECUTE], token, MAX_TOKEN);
	} else if (!strcmp (id, "ln"))
	{
		strncpy (tokens[LINK_OK], token, MAX_TOKEN);
	} else if (!strcmp (id, "or"))
	{
		strncpy (tokens[LINK_BAD], token, MAX_TOKEN);
	} else if (!strcmp (id, "bd"))
	{
		strncpy (tokens[BLOCK], token, MAX_TOKEN);
	} else if (!strcmp (id, "cd"))
	{
		strncpy (tokens[CHAR], token, MAX_TOKEN);
	} else if (!strcmp (id, "lc"))
	{
		strncpy (tokens[LEFTCODE], token, MAX_TOKEN);
	} else if (!strcmp (id, "rc"))
	{
		strncpy (tokens[RIGHTCODE], token, MAX_TOKEN);
	} else if (!strcmp (id, "no"))
	{
		strncpy (tokens[NORMAL], token, MAX_TOKEN);
	} else if (!strcmp (id, "pi"))
	{
		strncpy (tokens[FIFO], token, MAX_TOKEN);
	} else if (!strcmp (id, "so"))
	{
		strncpy (tokens[SOCKET], token, MAX_TOKEN);
	}
}

/*read and fill tokens from LS_COLORS variable*/
void
init_tokens (const char *str)
{
	char    id[MAX_TOKEN + 2];
	char   *id_p = id;

	char    token[MAX_TOKEN + 2];
	char   *token_p = token;

	int     in_id = 1, num = 0;

	if (str == NULL)
		return;

	for (; *str; str++)
	{
		switch (*str)
		{
			case '=':	/*end of id */
				in_id = 0;
				*id_p = 0;	/*finish string */
				num = 0;
				break;
			case ':':	/*end of token */
				in_id = 1;
				*token_p = 0;	/*finish string */
				num = 0;
				process_token (id, token);
				id_p = id;
				token_p = token;
				break;
			default:	/*collect strings */
				if (in_id)
				{
					*id_p = *str;
					if (num < MAX_TOKEN)
					{
						id_p++;
						num++;
					} else
					{
						puts ("init_tokens:ID size too big");
						exit (EXIT_FAILURE);
					}
				} else
				{
					*token_p = *str;
					if (num < MAX_TOKEN)
					{
						token_p++;
						num++;
					} else
					{
						puts ("init_tokens:TOKEN size too big");
						exit (EXIT_FAILURE);
					}
				}
				break;
		}
	}
}

/*setup colors*/
void
init_colors ()
{
	char   *ls_colors;

	if (force_color || isatty (STDOUT_FILENO))
	{
		if (((ls_colors = getenv ("LS_COLORS")) != NULL
		     && !force_no_color) || force_color)
		{
			is_colorize = 1;
			init_tokens (ls_colors);
		}
	}
}

/*colorize output*/
void
colorize (const char *message, int flag)
{
	if (is_colorize)
	{
		printf ("%s%s%s%s%s%s%s",
			tokens[LEFTCODE], tokens[flag], tokens[RIGHTCODE],
			message, tokens[LEFTCODE], tokens[NORMAL],
			tokens[RIGHTCODE]);
		return;
	}
	printf ("%s", message);
}

#define bufsiz 32

/*replace non-printable characters with '?"*/
char   *
printable_name (const char *base_name)
{
	wchar_t buf[bufsiz + 1];
	mbstate_t i_state, o_state;

	unsigned o_size = strlen(base_name) + 1;
	char *name = xmalloc(o_size);
	int name_size = 0; /* number of bytes written to name */

	memset(&i_state, 0, sizeof(i_state));
	memset(&o_state, 0, sizeof(o_state));

	/* mbsrtowcs will set base_name to NULL when the whole string
	   has been converted. */
	while(base_name)
	{
		size_t res, i, o_len;
		mbstate_t tmp;
		const wchar_t *buf_p = buf;

		res = mbsrtowcs(buf, &base_name, bufsiz, &i_state);
		if(res == (size_t)-1) {
			/* Invalid multibyte sequence.
			 * Well, garbage in -> garbage out. */
			xfree(name);
			return strdup("[invalid multibyte sequence]");
		}

		/* Now we have 'res' new wchars in buf;
		 * change non-printable ones to L'?'... */
		for(i = 0; i < res; i++)
		{
			if (!iswprint((wint_t) buf[i]))
			{
				buf[i] = L'?';
			}
		}

		/* and write the text to the output string. */
		tmp = o_state;
		o_len = wcsnrtombs(NULL, &buf_p, res, 0, &tmp);
		if(o_len == (size_t)-1)
		{
			xfree(name);
			return strdup("[conversion error]");
		}
		if(o_len > o_size - name_size - 1)
		{
			o_size += o_len * 2;
			name = xrealloc(name, o_size);
		}
		buf_p = buf;
		wcsnrtombs(name + name_size, &buf_p, res,
				o_len, &o_state);
		name_size += o_len;
	}

	name[name_size] = '\0';
	return name;
}

/*stat file and colorize it*/
void
print_stat (const char *__base_name, struct stat *st)
{
	char    buf[PATH_MAX];
	struct stat linkstat;
	int     res;
	char   *base_name = printable_name (__base_name);

	if (S_ISDIR (st->st_mode))
	{
		colorize (base_name, DIRECTORY);
	} else if (S_ISREG (st->st_mode))
	{
		if ((S_IXUSR & st->st_mode) ||
		    (S_IXGRP & st->st_mode) || (S_IXOTH & st->st_mode))
		{
			colorize (base_name, EXECUTE);
		} else
		{
			colorize (base_name, REGULAR);
		}
	} else if (S_ISLNK (st->st_mode))
	{
		res = readlink (base_name, buf, sizeof (buf));
		if (res < 0)
		{
			colorize (base_name, LINK_BAD);
			colorize (" - [unable to read link info]", LINK_BAD);
		} else
		{
			buf[res] = 0;
			if (stat (buf, &linkstat))
			{
				colorize (base_name, LINK_BAD);
				printf (" -> ");
				colorize (buf, LINK_BAD);
			} else
			{
				colorize (base_name, LINK_OK);
				printf (" -> ");
				print_stat (buf, &linkstat);
			}

		}
	} else if (S_ISCHR (st->st_mode))
	{
		colorize (base_name, CHAR);
	} else if (S_ISBLK (st->st_mode))
	{
		colorize (base_name, BLOCK);
	} else if (S_ISFIFO (st->st_mode))
	{
		colorize (base_name, FIFO);
	} else if (S_ISSOCK (st->st_mode))
	{
		colorize (base_name, SOCKET);
	} else
	{
		colorize (base_name, NORMAL);
	}

	free (base_name);
}

void
match_cwd_stat (struct stat *prev_stat)
{
	struct stat new_stat;

	if (lstat (".", &new_stat) < 0)
		error (EXIT_FAILURE, errno, "lstat('.')");
#if 0

	printf ("prev: dev %d ino %d\n", (int) prev_stat->st_dev,
		(int) prev_stat->st_ino);
	printf ("new: dev %d ino %d\n", (int) new_stat.st_dev,
		(int) new_stat.st_ino);
#endif

	if ((new_stat.st_dev != prev_stat->st_dev) ||
	    ((new_stat.st_ino != prev_stat->st_ino)))
		error (EXIT_FAILURE, 0, "wrong returning to up directory");
}

/*secure returning into a directory*/
void sec_return(const char *dest_path, struct stat *dest_stat)
{
	if (chdir (dest_path) < 0)
		error (EXIT_FAILURE, errno, "chdir(%s)", dest_path);
	match_cwd_stat(dest_stat);
}

/*filtering: return 1 if we must skip it*/
int
skip_it (const char *name, struct stat *st)
{
	mode_t  file_mode = st->st_mode;

	int     ret = (filter[(file_mode & S_IFMT) >> 12] == 0);

	if (filter_name)
		ret |= regexec (&re, name, 0, NULL, 0);

	return ret;
}

/*just check dir for skipped elements*/
int
check_dir ()
{
	DIR    *d = NULL;
	struct dirent *de = NULL;

	d = opendir (".");	/*open current directory */

	/*don't print unreadable directories */
	if (!d)
		return 1;	/*return skip */

	/*walk throw content of the directory */
	while ((de = readdir (d)) != NULL)
	{
		struct stat st;
		char    name[PATH_MAX];

		if (strlen (de->d_name) > PATH_MAX)
		{
			printf ("warning: name too long");
			return 1;
		}
		strncpy (name, de->d_name, PATH_MAX);

		/*skip standart subdirs */
		if (!strcmp (name, ".") || !strcmp (name, ".."))
			continue;

		if (lstat (name, &st) < 0)
			error (EXIT_FAILURE, errno, "lstat('%s')", name);

		if (!skip_it (name, &st))
		{
			closedir (d);
			return 0;	/*we have entry not to skip */
		}

		if (S_ISDIR (st.st_mode))
		{		/*return result from subdir */
			int     res = 1;	/*skip dirs we cannot chdir */
			struct stat saved_st;

			if (fstat (dirfd (d), &saved_st) < 0)
				error (EXIT_FAILURE, errno, "fstat(dirfd)");

			if (!chdir (name))
			{
				res = check_dir ();
				sec_return ("..", &saved_st);
			}

			if (!res)
			{	/*have non-skipped elements in subdir */
				closedir (d);
				return 0;
			}

		}
	}
	closedir (d);
	return 1;		/*skeep all subelement's */
}

/*skip processing with the subdirectory support*/
int
big_skip (const char *name, struct stat *st)
{
	int     skip;

	skip = skip_it (name, st);	/*skip or not element */
	if (S_ISDIR (st->st_mode))
	{			/*still print dir if it have unfiltered content */

		struct stat saved_st;

		if (lstat (".", &saved_st) < 0)
			error (EXIT_FAILURE, errno, "lstat('.')");

		if (!chdir (name))
		{
			skip &= check_dir ();	/*recursive checking */
			sec_return ("..", &saved_st);
		}
	}
	return skip;
}

void
process_dir (const char *prefix, int level)
{
	DIR    *d = NULL;
	struct dirent *de = NULL;

	/*break if we reach maximum directory level */
	if ((max_level >= 0) && (level > max_level))
		return;

	/*return if prefix is full */
	if (strlen (prefix) >= (PATH_MAX - 4))
	{
		fprintf (stderr, "warning: prefix too long\n");
		return;
	}

	d = opendir (".");	/*open current directory */

	/*skip unreadable directories */
	if (!d)
		return;

	/*walk throw content of the directory */
	while ((de = readdir (d)) != NULL)
	{
		struct stat st;
		off_t   curr_offset = telldir (d);
		int     skip = 0;	/*for filter */
		int     final = 0;
		char    name[PATH_MAX];	/*saved name */

		if (strlen (de->d_name) > PATH_MAX)
		{
			printf ("warning: name too long");
			closedir (d);
			return;
		}

		strncpy (name, de->d_name, PATH_MAX);

		/*skip standart subdirs */
		if (!strcmp (name, ".") || !strcmp (name, ".."))
			continue;

		/*fetch stat information */
		if (lstat (name, &st) < 0)
			error (EXIT_FAILURE, errno, "lstat('%s')", name);

		/*read forward:made prefix */
		final = (readdir (d) == NULL);
		seekdir (d, curr_offset);

		/*advanced check for final */
		if (is_filter && !final)
		{
			struct dirent *nn_de;
			struct stat nn_st;

			/*look forward to the end of dir */
			while ((nn_de = readdir (d)) != NULL)
			{
				if (lstat (nn_de->d_name, &nn_st) < 0)
					error (EXIT_FAILURE, errno,
					       "lstat('%s')", nn_de->d_name);
				if (!big_skip (nn_de->d_name, &nn_st))
				{	/*we have unskipped elements after current */
					break;
				}
			}
			seekdir (d, curr_offset);

			if (!nn_de)
			{
				final = 1;
			}
		}

		/*filter this item or not */
		if (is_filter)
		{
			skip = big_skip (name, &st);
		}

		/*print filename with colors */
		if (!skip)
		{
			printf ("%s%s%s%s", prefix,
				(final) ? gr->coner : gr->branch,
				gr->horizontal, gr->horizontal);
			print_stat (name, &st);
			puts ("");
		}

		/*recursive walking into subdirs */
		if (S_ISDIR (st.st_mode))
		{
			char   *new_prefix = xmalloc (PATH_MAX);
			struct stat saved_st;

			if (fstat (dirfd (d), &saved_st) < 0)
				error (EXIT_FAILURE, errno, "fstat(dirfd)");

			/*append three symbols to prefix */
			snprintf (new_prefix, PATH_MAX, "%s%s%s%s", prefix,
				  (final) ? gr->empty : gr->vertical,
				  gr->empty, gr->empty);

			/*FIXME: made it secure */
			if (!chdir (name))
			{
				process_dir (new_prefix, level + 1);

				sec_return ("..", &saved_st);
			}

			free (new_prefix);
		}

	}
	closedir (d);
	return;
}

void
print_tree (const char *name)
{
	struct stat st;

	init_colors ();

	if (lstat (name, &st))
		error (EXIT_FAILURE, errno, "lstat('%s')", name);

	print_stat (name, &st);
	puts ("");
	/*print tree */
	if (S_ISDIR (st.st_mode))
	{
		char *cwd_path = NULL;
		struct stat cwd_st;
		char *prefix = xmalloc (PATH_MAX);

		prefix[0] = 0;

		if (strcmp(name, "."))
		{
			cwd_path = get_current_dir_name();
			if (lstat (".", &cwd_st) < 0)
				error (EXIT_FAILURE, errno, "lstat('.')");
		}

		if (!chdir (name))
		{
			process_dir (prefix, 0);
			if (cwd_path)
			{
				sec_return (cwd_path, &cwd_st);
				free (cwd_path);
			}
		}
		free (prefix);
	}
}

void
usage (int retcode)
{
	fprintf (stdout,
		 "Usage: %s [-hvUGCnL] [--force-[no-]color] [<directory list>]\n",
		 __progname);
	fputs ("\nTree - list contents of directories in a tree-like format\n\n"
	    "With no arguments, tree lists the files in the current\n"
	    "directory. When directory arguments are given, tree lists all\n"
	    "the files and/or directories found in the given directories each in turn.\n\n",
	       stdout);

	fputs ("\t-h, --help             display help screen\n"
	       "\t-v, --version          display version information\n"
	       "\t-C, --force-color      force colorization even on non-tty outputs\n"
	       "\t-n, --force-no-color   force no colorization even on tty outputs\n"
	       "\t-L, --max-level   	  set maximum tree depth\n"
	       "\t-N, --name   	  print only entries matched to this name\n"
	       "\t-G, --graphics-vt100   use VT100 graphics to display tree\n"
	       "\t-U, --graphics-utf8    use UTF8  graphics to display tree\n"
	       "\t-T, --type <list>	  filter: display only types in list\n",
	       stdout);
	fputs ("\t\ttypes:\n" "\t\tf - regular files\n"
	       "\t\tl - symbolic links\n" "\t\tp - FIFOs\n"
	       "\t\ts - sockets\n"
	       "\t\td - directories (always enabled in tree mode)\n"
	       "\t\tb - block (buffered) special\n"
	       "\t\tc - character (unbuffered) special\n" "\t\te.g.:\n"
	       "\t\t--type fl will display only regular files and symlinks\n\n",
	       stdout);

	fputs ("By default, when a symbolic link is encountered, "
			"the path that the symbolic\n"
			"link refers to is printed after the name of the link "
			"in the format:\nname -> real-path\n", stdout);
	fputs ("\nReport bugs to <inger@altlinux.org>\n", stdout);

	exit (retcode);
}

int
main (int argc, char *argv[])
{
	int     c;

	setlocale (LC_ALL, "");

	while (1)
	{
		static struct option long_options[] = {
			{"help", no_argument, 0, 'h'},
			{"version", no_argument, 0, 'v'},
			/*visualisation */
			{"force-color", no_argument, 0, 'C'},
			{"force-no-color", no_argument, 0, 'n'},
			{"graphics-vt100", no_argument, 0, 'G'},
			{"graphics-utf8", no_argument, 0, 'U'},
			{"max-level", required_argument, 0, 'L'},
			/*{"plain", no_argument, 0, 'P'}, */
			/*filtering */
			{"type", required_argument, 0, 'T'},
			{"name", required_argument, 0, 'N'},
			{0, 0, 0, 0}
		};

		c = getopt_long (argc, argv, "hvCnGUL:T:N:", long_options,
				 NULL);
		if (c == -1)
			break;
		switch (c)
		{
			case 'h':
				usage (EXIT_SUCCESS);
			case 'v':
				fputs ("tree version " VERSION "\n"
				       "Written by Stanislav Ievlev\n\n"
				       "Copyright (C) 2002-2004 ALT Linux Team\n",
				       stdout);
				exit (EXIT_SUCCESS);
			case 'C':
				if (force_no_color)
				{
					fprintf (stderr,
						 "conflicting options\n");
					exit (EXIT_FAILURE);
				}
				force_color = 1;
				break;

			case 'n':
				if (force_color)
				{
					fprintf (stderr,
						 "conflicting options\n");
					exit (EXIT_FAILURE);
				}
				force_no_color = 1;
				break;
			case 'L':
				max_level = strtol (optarg, 0, 0);
				break;
			case 'T':
				filter_type = optarg;
				is_filter = 1;
				break;
			case 'N':
				filter_name = optarg;
				is_filter = 1;
				break;
			case 'G':
				gr = &sym_vt100;
				break;
			case 'U':
				gr = &sym_utf8;
				break;
				/*case 'P':
				   is_plain = 1;
				   break; */
			default:
				usage (EXIT_FAILURE);
		}
	}

	/*fill filter mask if we have */
	if (filter_type != NULL)
	{
		memset (filter, 0, sizeof (filter));
		for (; *filter_type; filter_type++)
		{
			switch (*filter_type)
			{
				case 'f':
					filter[S_IFREG >> 12] = 1;
					break;
				case 'b':
					filter[S_IFBLK >> 12] = 1;
					break;
				case 'c':
					filter[S_IFCHR >> 12] = 1;
					break;
				case 'p':
					filter[S_IFIFO >> 12] = 1;
					break;
				case 'l':
					filter[S_IFLNK >> 12] = 1;
					break;
				case 's':
					filter[S_IFSOCK >> 12] = 1;
					break;
				case 'd':
					filter[S_IFDIR >> 12] = 1;
					break;
				default:
					break;
			}
		}
	} else
	{
		/*display all types */
		memset (filter, 1, sizeof (filter));
	}

	/*setup regexp if we want to use it */
	if (filter_name != NULL)
	{
		int     err = 0;

		if ((err = regcomp (&re, filter_name, 1)) != 0)
		{
			size_t  len = regerror (err, &re, NULL, 0);
			char   *buffer = xmalloc (len);

			(void) regerror (err, &re, buffer, len);
			fprintf (stderr, "unable to compile regexp:%s\n",
				 buffer);
			free (buffer);
			exit (EXIT_FAILURE);
		}
	}

	init_colors ();

	/* Restore default color before exiting */
	signal (SIGINT, restore_default_colors_handler);
	signal (SIGTERM, restore_default_colors_handler);
	signal (SIGQUIT, restore_default_colors_handler);
	signal (SIGTSTP, sigtstp_handler);

	if (optind < argc)
	{
		while (optind < argc)
		{
			print_tree (argv[optind++]);
		}
	} else
	{
		print_tree (".");
	}

	restore_default_colors ();

	return 0;
}
